class Layerpedia:
    layer_mapping_dict = {
        'Activation': {'is_trainable': False, 'kernel_index': None},
        'ActivityRegularization': {'is_trainable': False, 'kernel_index': None},
        'Add': {'is_trainable': False, 'kernel_index': None},
        'AdditiveAttention': {'is_trainable': False, 'kernel_index': None},
        'Attention': {'is_trainable': False, 'kernel_index': None},
        'Average': {'is_trainable': False, 'kernel_index': None},
        'AveragePooling1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'AveragePooling2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'AveragePooling3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'BatchNormalization': {'is_trainable': False, 'kernel_index': None},
        'Concatenate': {'is_trainable': False, 'kernel_index': None},
        'Conv1D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 2, "rank": 1, 'enable_bigger_input_rank': True},
        'Conv2D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 3, "rank": 2, 'enable_bigger_input_rank': True},
        'Conv3D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 4, "rank": 3, 'enable_bigger_input_rank': True},
        'Conv1DTranspose': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 1, "rank": 1, 'enable_bigger_input_rank': True},
        'Conv2DTranspose': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 2, "rank": 2, 'enable_bigger_input_rank': True},
        'Conv3DTranspose': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 3, "rank": 3, 'enable_bigger_input_rank': True},
        'ConvLSTM1D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': None, "rank": 1, 'enable_bigger_input_rank': True},
        'ConvLSTM2D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': None, "rank": 2, 'enable_bigger_input_rank': True},
        'ConvLSTM3D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': None, "rank": 3, 'enable_bigger_input_rank': True},
        'SeparableConv1D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 2, "rank": 1, 'enable_bigger_input_rank': True},
        'SeparableConv2D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 3, "rank": 2, 'enable_bigger_input_rank': True},
        'SeparableConv3D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 4, "rank": 3, 'enable_bigger_input_rank': True},
        'DepthwiseConv1D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 2, "rank": 1, 'enable_bigger_input_rank': True},  # TODO: review
        'DepthwiseConv2D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 3, "rank": 2, 'enable_bigger_input_rank': True},  # TODO: review
        'DepthwiseConv3D': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 4, "rank": 3, 'enable_bigger_input_rank': True},  # TODO: review
        'Cropping1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'Cropping2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'Cropping3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'Dense': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': 1},
        'Dot': {'is_trainable': False, 'kernel_index': None},
        'Dropout': {'is_trainable': False, 'kernel_index': None},
        'ELU': {'is_trainable': False, 'kernel_index': None},
        'Embedding': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': 0},
        'Flatten': {'is_trainable': False, 'kernel_index': None},
        'GRU': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': 1},
        'GRUCell': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': None},
        'GlobalAveragePooling1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'GlobalAveragePooling2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'GlobalAveragePooling3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'GlobalMaxPooling1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'GlobalMaxPooling2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'GlobalMaxPooling3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'LSTM': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': 1},
        'LSTMCell': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': None},
        'LayerNormalization': {'is_trainable': False, 'kernel_index': None},
        'LeakyReLU': {'is_trainable': False, 'kernel_index': None},
        'Masking': {'is_trainable': False, 'kernel_index': None},
        'MaxPooling1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'MaxPooling2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'MaxPooling3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'Maximum': {'is_trainable': False, 'kernel_index': None},
        'Minimum': {'is_trainable': False, 'kernel_index': None},
        'Multiply': {'is_trainable': False, 'kernel_index': None},
        'PReLU': {'is_trainable': False, 'kernel_index': None},
        'Permute': {'is_trainable': False, 'kernel_index': None},
        'RNN': {'is_trainable': False, 'kernel_index': None},
        'ReLU': {'is_trainable': False, 'kernel_index': None},
        'RepeatVector': {'is_trainable': False, 'kernel_index': None},
        'Reshape': {'is_trainable': False, 'kernel_index': None},
        'SimpleRNN': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': None},
        'SimpleRNNCell': {'is_trainable': True, 'filters_attr_name': 'units',  'kernel_index': None},
        'Softmax': {'is_trainable': False, 'kernel_index': None},
        'SpatialDropout1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'SpatialDropout2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'SpatialDropout3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'StackedRNNCells': {'is_trainable': False, 'kernel_index': None},
        'Subtract': {'is_trainable': True, 'filters_attr_name': 'filters',  'kernel_index': None},
        'ThresholdedReLU': {'is_trainable': False, 'kernel_index': None},
        'UpSampling1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'UpSampling2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'UpSampling3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'ZeroPadding1D': {'is_trainable': False, 'kernel_index': None, "rank": 1},
        'ZeroPadding2D': {'is_trainable': False, 'kernel_index': None, "rank": 2},
        'ZeroPadding3D': {'is_trainable': False, 'kernel_index': None, "rank": 3},
        'BinaryCrossentropy': {'is_trainable': False, 'kernel_index': None},
        'CategoricalCrossentropy': {'is_trainable': False, 'kernel_index': None},
        'CustomLoss': {'is_trainable': False, 'kernel_index': None},
        'CategoricalHinge': {'is_trainable': False, 'kernel_index': None},
        'CosineSimilarity': {'is_trainable': False, 'kernel_index': None},
        'Hinge': {'is_trainable': False, 'kernel_index': None},
        'Huber': {'is_trainable': False, 'kernel_index': None},
        'KLDivergence': {'is_trainable': False, 'kernel_index': None},
        'LogCosh': {'is_trainable': False, 'kernel_index': None},
        'MeanAbsoluteError': {'is_trainable': False, 'kernel_index': None},
        'MeanAbsolutePercentageError': {'is_trainable': False, 'kernel_index': None},
        'MeanSquaredError': {'is_trainable': False, 'kernel_index': None},
        'MeanSquaredLogarithmicError': {'is_trainable': False, 'kernel_index': None},
        'Poisson': {'is_trainable': False, 'kernel_index': None},
        'Reduction': {'is_trainable': False, 'kernel_index': None},
        'SparseCategoricalCrossentropy': {'is_trainable': False, 'kernel_index': None},
        'SquaredHinge': {'is_trainable': False, 'kernel_index': None},
        'Adadelta': {'is_trainable': False, 'kernel_index': None},
        'Adagrad': {'is_trainable': False, 'kernel_index': None},
        'Adam': {'is_trainable': False, 'kernel_index': None},
        'Adamax': {'is_trainable': False, 'kernel_index': None},
        'Ftrl': {'is_trainable': False, 'kernel_index': None},
        'Nadam': {'is_trainable': False, 'kernel_index': None},
        'RMSprop': {'is_trainable': False, 'kernel_index': None},
        'SGD': {'is_trainable': False, 'kernel_index': None},
        'Trainer': {'is_trainable': False, 'kernel_index': None},
        'Repeat': {'is_trainable': False, 'kernel_index': None},
        'MultiHeadAttention': {'is_trainable': True, 'filters_attr_name': '_key_dim', 'kernel_index': 2},
        'Gather': {'is_trainable': False, 'kernel_index': None},
        'Variable': {'is_trainable': True, 'kernel_index': None}
    }

    @staticmethod
    def get_layer_knowledge(layer_name: str):
        return Layerpedia.layer_mapping_dict.get(layer_name)



from typing import List, Set


def generate_nodes_data_file(dist: str, all_nodes, args):
    all_content: List[str] = []
    imports: Set[str] = set()
    enums_cache: Set[str] = set()
    enums: List[str] = []
    nodes_data: List[str] = []
    node_names: List[str] = []

    for node in all_nodes:
        imports.add('from dataclasses import dataclass\n')
        node_name = node['name'].replace(' ', '')
        node_names.append(node_name)
        nodes_data.append(f"@dataclass\nclass {node_name}:\n")
        node_props = node['properties']

        for prop in node_props:
            p_type = handle_prop(enums_cache, enums, imports, args, node_name, prop)
            nodes_data.append(f"{tab}{prop['name']}: {p_type}\n")

        nodes_data.append(f"{tab}type: NodeType\n\n\n")

    enums.append(f"class NodeType(Enum):\n")
    for e_val in node_names:
        enums.append('%s%s = "%s"\n' % (tab, e_val, e_val))
    enums.append(f"\n\n")

    complex_node_data_type = generate_complex_type(imports, node_names)

    all_content.extend(imports)
    all_content.append('\n\n')
    all_content.extend(enums)
    all_content.extend(nodes_data)
    all_content.append(complex_node_data_type)

    full_text = ''.join(all_content)
    with open(f'{dist}nodedata.py', "w") as p_file:
        p_file.write(full_text)


def generate_complex_type(imports: Set[str], node_names: List[str]) -> str:
    imports.add('from typing import Union\n')
    lines: List[str] = ['@dataclass\nclass NodeDataTypes:\n']
    c_line = f"{tab}props: Union["
    for name in node_names:
        if len(c_line) + len(name) < line_length:
            c_line = f"{c_line}{name}, "
        else:
            lines.append(f"{c_line}\n")
            c_line = f"{union_prop_tab}{name}, "
    lines.append(f"{c_line[0: len(c_line) - 2]}]\n")
    return ''.join(lines)


def handle_prop(enums_cache: Set[str], enums: List[str], imports: Set[str],
                args, node_name: str, prop: dict) -> str:
    if prop['type'] == 'select':
        if prop['name'] in args['arg_group']:
            p_name = args['arg_group'][prop['name']]
            enum_values = args['arg_values'][p_name]
        else:
            p_name = prop['name']
            enum_values = prop['values']
        p_type = ''.join(word.title() for word in p_name.split('_')) + 'Enum'
        if p_name not in enums_cache:
            imports.add('from enum import Enum\n')
            enums_cache.add(p_name)
            enums.append(f"class {p_type}(Enum):\n")
            for e_val in enum_values:
                enums.append('%s%s = "%s"\n' % (tab, e_val, e_val))
            enums.append("\n\n")
        return p_type

    if prop['type'] == 'NoneType':
        return 'int' if is_one_dimension(node_name) else 'List[int]'

    if prop['type'] == 'tuple':
        imports.add('from typing import List\n')
        return 'List[int]'

    return prop['type']


def is_one_dimension(node_name):
    return '1D' in node_name


tab = ' ' * 4
union_prop_tab = ' ' * 17
line_length = 120



import glob
import json
import ntpath
import tarfile
import tempfile
from importlib.util import find_spec
from pathlib import Path
from typing import Callable, Optional, List, Dict, Tuple, Type

import tensorflow as tf  # type: ignore
from keras import Model  # type: ignore
from keras_data_format_converter import convert_channels_first_to_last  # type: ignore
from leap_model_rebuilder import rebuild_model  # type: ignore
from onnx2kerastl import onnx_to_keras  # type: ignore
from onnx2kerastl.exceptions import OnnxUnsupported  # type: ignore
from tensorflow.keras.models import load_model  # type: ignore

from leap_model_parser.contract.importmodelresponse import InputInfo, NodeResponse, ImportModelTypeEnum
from leap_model_parser.keras_json_model_import import KerasJsonModelImport

onnx_imported = False
package_name = 'onnx'
spec = find_spec(package_name)
if spec is not None:
    import onnx  # type: ignore

    onnx_imported = True


class ModelParser:
    def __init__(self, should_transform_inputs_and_outputs=False,
                 custom_layers: Optional[Dict[str, Type[tf.keras.layers.Layer]]] = None):
        self._should_transform_inputs_and_outputs = should_transform_inputs_and_outputs
        self.custom_layers = custom_layers
        if custom_layers is None:
            self.custom_layers = {}

        self._model_types_converter = {
            ImportModelTypeEnum.JSON_TF2.value: self.convert_json_model,
            ImportModelTypeEnum.H5_TF2.value: self.convert_h5_model,
            ImportModelTypeEnum.ONNX.value: self.convert_onnx_model,
            ImportModelTypeEnum.PB_TF2.value: self.convert_pb_model,
        }

    def get_keras_model_and_model_graph(
            self, model_path: Path, model_type: ImportModelTypeEnum) -> Tuple[Dict[str, NodeResponse], List[InputInfo], Optional[Model]]:
        model_to_keras_converter: Optional[Callable[[str], Tuple[Dict[str, NodeResponse], Model]]] = \
            self._model_types_converter.get(model_type.value)
        if model_to_keras_converter is None:
            raise Exception(
                f"Unable to import external version, {str(model_path)} file format isn't supported")

        file_path = str(model_path)
        try:
            model_schema, keras_model_with_weights = model_to_keras_converter(file_path)
            model_generator = KerasJsonModelImport(self.custom_layers)

            keras_model = keras_model_with_weights
            if keras_model is None:
                keras_model = tf.keras.models.model_from_json(json.dumps(model_schema))
            layer_name_to_inbound_nodes = {
                layer.name: layer.inbound_nodes
                for layer in keras_model.layers
            }

            graph, connected_inputs = model_generator.generate_graph(
                model_schema, layer_name_to_inbound_nodes)
            return graph, connected_inputs, keras_model_with_weights
        except Exception as e:
            if model_type.value in [ImportModelTypeEnum.H5_TF2.value, ImportModelTypeEnum.PB_TF2.value]:
                if model_type.value == ImportModelTypeEnum.H5_TF2.value:
                    keras_model = self._load_keras_model_with_custom_layers(
                        file_path)
                else:
                    keras_model = self._get_k_model_from_pb_path(file_path)

                rebuilt_model = rebuild_model(keras_model)
                model_schema, keras_model = self.convert_to_keras_model(
                    rebuilt_model)

                layer_name_to_inbound_nodes = {
                    layer.name: layer.inbound_nodes
                    for layer in keras_model.layers
                }
                model_generator = KerasJsonModelImport(self.custom_layers)
                graph, connected_inputs = model_generator.generate_graph(
                    model_schema, layer_name_to_inbound_nodes)
                return graph, connected_inputs, keras_model
            else:
                raise e

    def _get_k_model_from_pb_path(self, file_path: str):
        tar_file = tarfile.open(file_path)
        with tempfile.TemporaryDirectory() as temp_dir:
            tar_file.extractall(temp_dir)
            pb_files = glob.glob(temp_dir + "/**/*.pb", recursive=True)
            if len(pb_files) == 0:
                raise Exception('no pb files were found')

            pb_file_path = next(iter(pb_files))
            pb_folder_path = next(iter(ntpath.split(pb_file_path)))

            k_model = self._load_keras_model_with_custom_layers(pb_folder_path)
        return k_model

    def generate_model_graph(self, model_path: Path, model_type: ImportModelTypeEnum) -> Tuple[Dict[str, NodeResponse], List[InputInfo]]:
        model_graph, connected_inputs, _ = self.get_keras_model_and_model_graph(
            model_path, model_type)
        return model_graph, connected_inputs

    @classmethod
    def convert_json_model(cls, file_path: str) -> Tuple[Dict[str, NodeResponse], None]:
        with open(file_path, 'r') as f:
            model_schema = json.load(f)
        return model_schema, None

    def convert_pb_model(self, file_path: str) -> Tuple[Dict[str, NodeResponse], Model]:
        k_model = self._get_k_model_from_pb_path(file_path)
        return self.convert_to_keras_model(k_model)

    def convert_onnx_model(self, file_path: str) -> Tuple[Dict[str, NodeResponse], Model]:
        if not onnx_imported:
            raise OnnxUnsupported()

        onnx_model = onnx.load_model(file_path)
        input_all = [_input.name for _input in onnx_model.graph.input]
        input_initializer = [
            node.name for node in onnx_model.graph.initializer]
        input_names = list(set(input_all) - set(input_initializer))
        k_model = onnx_to_keras(onnx_model, input_names=input_names,
                                name_policy='attach_weights_name')
        return self.convert_to_keras_model(k_model)

    def _load_keras_model_with_custom_layers(self, file_path: str):
        custom_objects = {}
        if self.custom_layers is not None:
            custom_objects = self.custom_layers
        return load_model(file_path, custom_objects=custom_objects, compile=False)

    def convert_h5_model(self, file_path: str) -> Tuple[Dict[str, NodeResponse], Model]:
        imported_model = self._load_keras_model_with_custom_layers(file_path)
        return self.convert_to_keras_model(imported_model)

    def convert_to_keras_model(self, k_model) -> Tuple[Dict[str, NodeResponse], Model]:
        converted_k_model = convert_channels_first_to_last(
            k_model, self._should_transform_inputs_and_outputs, self.custom_layers)

        model_schema = json.loads(converted_k_model.to_json())

        return model_schema, converted_k_model



import inspect
from types import FunctionType


class LeapInspect:
    @staticmethod
    def get_classes(cls, filterbyparents=None):
        """

        :param cls:
        :return:
        [
            {
                name: x
                class: x
                parents: []
            }
        ]
        """
        classes = []
        for c in inspect.getmembers(cls, predicate=inspect.isclass):
            if '__module__' not in c[1].__dict__:
                print('')
                continue
            cls_dict = {
                'name': c[0],
                'class': c[1],
                'parents': c[1].__dict__['__module__'].split('.')
            }
            if 'meta' in cls_dict['parents']: continue

            if len(set(filterbyparents) - set(cls_dict['parents'])) == len(set(filterbyparents)):
                classes.append(cls_dict)

        return classes

    @staticmethod
    def get_type(func, pval):
        dtype_map = {'int': ['int', 'Integer', 'number'], 'float': ['float'], 'str': ['string', 'str'],
                     'name': ['name'], 'tuple': ["tuple", "Tuple"]}
        if pval.default != inspect._empty:
            return str(type(pval.default)).split('\'')[1]
        # TODO: comment out due to missing CallableMeta type in python 3.8, if an error occurs from this section consider
        # finding the right way to write this if in python 3.8
        # if isinstance(pval.annotation, CallableMeta):
        #     return Callable
        if pval.annotation is str:
            return str
        _type = 'None'
        doc_str = func.__dict__.get('__doc__')
        if doc_str is not None:
            for line in doc_str.split('\n'):
                if ' {}:'.format(pval.name) in line[0:len(pval.name) + 10]:
                    # _type = line.split(': ')[1].split(',')[0].split(' ')[0]
                    for t, vals in dtype_map.items():
                        if True in list(val in line for val in vals):
                            _type = t
                    break
        return _type

    @staticmethod
    def get_function_args(func):
        args = {}
        if isinstance(func, FunctionType):
            parameters = inspect.signature(func).parameters
        else:
            parameters = inspect.signature(func.__init__).parameters
        for p, pval in parameters.items():
            if p in ['kwargs']:
                continue
            _type = LeapInspect.get_type(func, pval)
            args[pval.name] = \
                {
                    'name': pval.name,
                    'isdefault': pval.default != inspect._empty,
                    'default_val': pval.default if pval.default != inspect._empty else None,
                    'type': _type,
                }
        return args



from enum import Enum
from typing import List
from typing import Union
from dataclasses import dataclass


class ActivationsEnum(Enum):
    relu = "relu"
    softmax = "softmax"
    selu = "selu"
    softplus = "softplus"
    softsign = "softsign"
    tanh = "tanh"
    sigmoid = "sigmoid"
    hard_sigmoid = "hard_sigmoid"
    exponential = "exponential"
    linear = "linear"
    elu = "elu"
    swish = "swish"


class PaddingEnum(Enum):
    valid = "valid"
    same = "same"


class InitializersEnum(Enum):
    glorot_uniform = "glorot_uniform"
    glorot_normal = "glorot_normal"
    lecun_uniform = "lecun_uniform"
    he_normal = "he_normal"
    lecun_normal = "lecun_normal"
    he_uniform = "he_uniform"
    Identity = "Identity"
    Orthogonal = "Orthogonal"
    Zeros = "Zeros"
    Ones = "Ones"
    RandomNormal = "RandomNormal"
    RandomUniform = "RandomUniform"


class RegularizerEnum(Enum):
    L1 = "L1"
    L1L2 = "L1L2"
    L2 = "L2"


class ConstraintEnum(Enum):
    MaxNorm = "MaxNorm"
    MinMaxNorm = "MinMaxNorm"
    NonNeg = "NonNeg"
    RadialConstraint = "RadialConstraint"
    UnitNorm = "UnitNorm"


class InitializerEnum(Enum):
    glorot_uniform = "glorot_uniform"
    glorot_normal = "glorot_normal"
    lecun_uniform = "lecun_uniform"
    he_normal = "he_normal"
    lecun_normal = "lecun_normal"
    he_uniform = "he_uniform"
    Identity = "Identity"
    Orthogonal = "Orthogonal"
    Zeros = "Zeros"
    Ones = "Ones"
    RandomNormal = "RandomNormal"
    RandomUniform = "RandomUniform"


class NodeType(Enum):
    Activation = "Activation"
    ActivityRegularization = "ActivityRegularization"
    Add = "Add"
    AdditiveAttention = "AdditiveAttention"
    AlphaDropout = "AlphaDropout"
    Average = "Average"
    AveragePooling1D = "AveragePooling1D"
    AveragePooling2D = "AveragePooling2D"
    AveragePooling3D = "AveragePooling3D"
    BatchNormalization = "BatchNormalization"
    Bidirectional = "Bidirectional"
    CategoryEncoding = "CategoryEncoding"
    CenterCrop = "CenterCrop"
    Concatenate = "Concatenate"
    Conv1D = "Conv1D"
    Conv1DTranspose = "Conv1DTranspose"
    Conv2D = "Conv2D"
    Conv2DTranspose = "Conv2DTranspose"
    Conv3D = "Conv3D"
    Conv3DTranspose = "Conv3DTranspose"
    ConvLSTM1D = "ConvLSTM1D"
    ConvLSTM2D = "ConvLSTM2D"
    ConvLSTM3D = "ConvLSTM3D"
    Cropping1D = "Cropping1D"
    Cropping2D = "Cropping2D"
    Cropping3D = "Cropping3D"
    CuDNNGRU = "CuDNNGRU"
    CuDNNLSTM = "CuDNNLSTM"
    Dense = "Dense"
    DepthwiseConv1D = "DepthwiseConv1D"
    DepthwiseConv2D = "DepthwiseConv2D"
    Discretization = "Discretization"
    Dot = "Dot"
    Dropout = "Dropout"
    ELU = "ELU"
    Embedding = "Embedding"
    Flatten = "Flatten"
    GRU = "GRU"
    GRUCell = "GRUCell"
    GRUCellV1 = "GRUCellV1"
    GRUCellV2 = "GRUCellV2"
    GRUV1 = "GRUV1"
    GRUV2 = "GRUV2"
    GaussianDropout = "GaussianDropout"
    GaussianNoise = "GaussianNoise"
    GlobalAveragePooling1D = "GlobalAveragePooling1D"
    GlobalAveragePooling2D = "GlobalAveragePooling2D"
    GlobalAveragePooling3D = "GlobalAveragePooling3D"
    GlobalMaxPooling1D = "GlobalMaxPooling1D"
    GlobalMaxPooling2D = "GlobalMaxPooling2D"
    GlobalMaxPooling3D = "GlobalMaxPooling3D"
    GroupNormalization = "GroupNormalization"
    IntegerLookup = "IntegerLookup"
    LSTM = "LSTM"
    LSTMCell = "LSTMCell"
    LSTMCellV1 = "LSTMCellV1"
    LSTMCellV2 = "LSTMCellV2"
    LSTMV1 = "LSTMV1"
    LSTMV2 = "LSTMV2"
    LayerNormalization = "LayerNormalization"
    LeakyReLU = "LeakyReLU"
    LocallyConnected1D = "LocallyConnected1D"
    LocallyConnected2D = "LocallyConnected2D"
    Masking = "Masking"
    MaxPooling1D = "MaxPooling1D"
    MaxPooling2D = "MaxPooling2D"
    MaxPooling3D = "MaxPooling3D"
    Maximum = "Maximum"
    Minimum = "Minimum"
    MultiHeadAttention = "MultiHeadAttention"
    Multiply = "Multiply"
    Normalization = "Normalization"
    PReLU = "PReLU"
    Permute = "Permute"
    RandomBrightness = "RandomBrightness"
    RandomContrast = "RandomContrast"
    RandomCrop = "RandomCrop"
    RandomFlip = "RandomFlip"
    RandomFourierFeatures = "RandomFourierFeatures"
    RandomHeight = "RandomHeight"
    RandomRotation = "RandomRotation"
    RandomTranslation = "RandomTranslation"
    RandomWidth = "RandomWidth"
    RandomZoom = "RandomZoom"
    ReLU = "ReLU"
    RepeatVector = "RepeatVector"
    Reshape = "Reshape"
    Resizing = "Resizing"
    SeparableConv1D = "SeparableConv1D"
    SeparableConv2D = "SeparableConv2D"
    SimpleRNN = "SimpleRNN"
    SimpleRNNCell = "SimpleRNNCell"
    Softmax = "Softmax"
    SpatialDropout1D = "SpatialDropout1D"
    SpatialDropout2D = "SpatialDropout2D"
    SpatialDropout3D = "SpatialDropout3D"
    StringLookup = "StringLookup"
    Subtract = "Subtract"
    SyncBatchNormalization = "SyncBatchNormalization"
    TextVectorization = "TextVectorization"
    ThresholdedReLU = "ThresholdedReLU"
    TimeDistributed = "TimeDistributed"
    UnitNormalization = "UnitNormalization"
    UpSampling1D = "UpSampling1D"
    UpSampling2D = "UpSampling2D"
    UpSampling3D = "UpSampling3D"
    ZeroPadding1D = "ZeroPadding1D"
    ZeroPadding2D = "ZeroPadding2D"
    ZeroPadding3D = "ZeroPadding3D"
    BinaryCrossentropy = "BinaryCrossentropy"
    BinaryFocalCrossentropy = "BinaryFocalCrossentropy"
    CategoricalCrossentropy = "CategoricalCrossentropy"
    CategoricalHinge = "CategoricalHinge"
    CosineSimilarity = "CosineSimilarity"
    Hinge = "Hinge"
    Huber = "Huber"
    KLDivergence = "KLDivergence"
    LogCosh = "LogCosh"
    MeanAbsoluteError = "MeanAbsoluteError"
    MeanAbsolutePercentageError = "MeanAbsolutePercentageError"
    MeanSquaredError = "MeanSquaredError"
    MeanSquaredLogarithmicError = "MeanSquaredLogarithmicError"
    Poisson = "Poisson"
    SquaredHinge = "SquaredHinge"
    Adadelta = "Adadelta"
    Adagrad = "Adagrad"
    Adam = "Adam"
    Adamax = "Adamax"
    Ftrl = "Ftrl"
    Nadam = "Nadam"
    RMSprop = "RMSprop"
    SGD = "SGD"
    OnnxAbs = "OnnxAbs"
    OnnxErf = "OnnxErf"
    OnnxHardSigmoid = "OnnxHardSigmoid"
    OnnxReduceMean = "OnnxReduceMean"
    OnnxSqrt = "OnnxSqrt"
    Dataset = "Dataset"
    Input = "Input"
    RepresentationBlock = "RepresentationBlock"
    GroundTruth = "GroundTruth"
    CustomLayer = "CustomLayer"
    CustomLoss = "CustomLoss"
    Visualizer = "Visualizer"
    Metric = "Metric"
    Lambda = "Lambda"
    TFOpLambda = "TFOpLambda"
    SlicingOpLambda = "SlicingOpLambda"
    Repeat = "Repeat"
    Variable = "Variable"
    Gather = "Gather"
    FixedDropout = "FixedDropout"


@dataclass
class Activation:
    activation: ActivationsEnum
    type: NodeType


@dataclass
class ActivityRegularization:
    l1: float
    l2: float
    type: NodeType


@dataclass
class Add:
    type: NodeType


@dataclass
class AdditiveAttention:
    use_scale: bool
    type: NodeType


@dataclass
class AlphaDropout:
    rate: float
    noise_shape: List[int]
    seed: List[int]
    type: NodeType


@dataclass
class Average:
    type: NodeType


@dataclass
class AveragePooling1D:
    pool_size: int
    strides: int
    padding: PaddingEnum
    type: NodeType


@dataclass
class AveragePooling2D:
    pool_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    type: NodeType


@dataclass
class AveragePooling3D:
    pool_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    type: NodeType


@dataclass
class BatchNormalization:
    axis: int
    momentum: float
    epsilon: float
    center: bool
    scale: bool
    beta_initializer: InitializersEnum
    gamma_initializer: InitializersEnum
    moving_mean_initializer: InitializersEnum
    moving_variance_initializer: InitializersEnum
    beta_regularizer: RegularizerEnum
    gamma_regularizer: RegularizerEnum
    beta_constraint: ConstraintEnum
    gamma_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Bidirectional:
    merge_mode: str
    weights: List[int]
    backward_layer: List[int]
    type: NodeType


@dataclass
class CategoryEncoding:
    num_tokens: List[int]
    output_mode: str
    sparse: bool
    type: NodeType


@dataclass
class CenterCrop:
    height: int
    width: int
    type: NodeType


@dataclass
class Concatenate:
    axis: int
    type: NodeType


@dataclass
class Conv1D:
    filters: int
    kernel_size: List[int]
    strides: int
    padding: PaddingEnum
    dilation_rate: int
    groups: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Conv1DTranspose:
    filters: int
    kernel_size: int
    strides: int
    padding: PaddingEnum
    output_padding: int
    dilation_rate: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Conv2D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    dilation_rate: List[int]
    groups: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Conv2DTranspose:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    output_padding: List[int]
    dilation_rate: List[int]
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Conv3D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    dilation_rate: List[int]
    groups: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Conv3DTranspose:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    output_padding: List[int]
    dilation_rate: List[int]
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class ConvLSTM1D:
    filters: int
    kernel_size: List[int]
    strides: int
    padding: PaddingEnum
    dilation_rate: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class ConvLSTM2D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    dilation_rate: List[int]
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class ConvLSTM3D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    dilation_rate: List[int]
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class Cropping1D:
    cropping: List[int]
    type: NodeType


@dataclass
class Cropping2D:
    cropping: List[int]
    type: NodeType


@dataclass
class Cropping3D:
    cropping: List[int]
    type: NodeType


@dataclass
class CuDNNGRU:
    units: int
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    type: NodeType


@dataclass
class CuDNNLSTM:
    units: int
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    type: NodeType


@dataclass
class Dense:
    units: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class DepthwiseConv1D:
    kernel_size: int
    strides: int
    padding: PaddingEnum
    depth_multiplier: int
    dilation_rate: int
    activation: ActivationsEnum
    use_bias: bool
    depthwise_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    depthwise_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    depthwise_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class DepthwiseConv2D:
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    depth_multiplier: int
    dilation_rate: List[int]
    activation: ActivationsEnum
    use_bias: bool
    depthwise_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    depthwise_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    depthwise_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Discretization:
    bin_boundaries: List[int]
    num_bins: List[int]
    epsilon: float
    output_mode: str
    sparse: bool
    type: NodeType


@dataclass
class Dot:
    axes: List[int]
    normalize: bool
    type: NodeType


@dataclass
class Dropout:
    rate: float
    noise_shape: List[int]
    seed: List[int]
    type: NodeType


@dataclass
class ELU:
    alpha: float
    type: NodeType


@dataclass
class Embedding:
    input_dim: int
    output_dim: int
    embeddings_initializer: str
    embeddings_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    embeddings_constraint: ConstraintEnum
    mask_zero: bool
    input_length: List[int]
    type: NodeType


@dataclass
class Flatten:
    type: NodeType


@dataclass
class GRU:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    unroll: bool
    time_major: bool
    reset_after: bool
    type: NodeType


@dataclass
class GRUCell:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    reset_after: bool
    type: NodeType


@dataclass
class GRUCellV1:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    reset_after: bool
    type: NodeType


@dataclass
class GRUCellV2:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    reset_after: bool
    type: NodeType


@dataclass
class GRUV1:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    unroll: bool
    reset_after: bool
    type: NodeType


@dataclass
class GRUV2:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    unroll: bool
    time_major: bool
    reset_after: bool
    type: NodeType


@dataclass
class GaussianDropout:
    rate: float
    seed: List[int]
    type: NodeType


@dataclass
class GaussianNoise:
    stddev: str
    seed: List[int]
    type: NodeType


@dataclass
class GlobalAveragePooling1D:
    type: NodeType


@dataclass
class GlobalAveragePooling2D:
    keepdims: bool
    type: NodeType


@dataclass
class GlobalAveragePooling3D:
    keepdims: bool
    type: NodeType


@dataclass
class GlobalMaxPooling1D:
    keepdims: bool
    type: NodeType


@dataclass
class GlobalMaxPooling2D:
    keepdims: bool
    type: NodeType


@dataclass
class GlobalMaxPooling3D:
    keepdims: bool
    type: NodeType


@dataclass
class GroupNormalization:
    groups: int
    axis: int
    epsilon: float
    center: bool
    scale: bool
    beta_initializer: InitializersEnum
    gamma_initializer: InitializersEnum
    beta_regularizer: RegularizerEnum
    gamma_regularizer: RegularizerEnum
    beta_constraint: ConstraintEnum
    gamma_constraint: ConstraintEnum
    type: NodeType


@dataclass
class IntegerLookup:
    max_tokens: List[int]
    num_oov_indices: int
    mask_token: List[int]
    oov_token: int
    vocabulary: List[int]
    vocabulary_dtype: str
    idf_weights: List[int]
    invert: bool
    output_mode: str
    sparse: bool
    pad_to_max_tokens: bool
    type: NodeType


@dataclass
class LSTM:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    time_major: bool
    unroll: bool
    type: NodeType


@dataclass
class LSTMCell:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class LSTMCellV1:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class LSTMCellV2:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class LSTMV1:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    unroll: bool
    type: NodeType


@dataclass
class LSTMV2:
    units: int
    activation: ActivationsEnum
    recurrent_activation: str
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    unit_forget_bias: bool
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    time_major: bool
    unroll: bool
    type: NodeType


@dataclass
class LayerNormalization:
    axis: int
    epsilon: float
    center: bool
    scale: bool
    beta_initializer: InitializersEnum
    gamma_initializer: InitializersEnum
    beta_regularizer: RegularizerEnum
    gamma_regularizer: RegularizerEnum
    beta_constraint: ConstraintEnum
    gamma_constraint: ConstraintEnum
    type: NodeType


@dataclass
class LeakyReLU:
    alpha: float
    type: NodeType


@dataclass
class LocallyConnected1D:
    filters: int
    kernel_size: List[int]
    strides: int
    padding: PaddingEnum
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    implementation: int
    type: NodeType


@dataclass
class LocallyConnected2D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    implementation: int
    type: NodeType


@dataclass
class Masking:
    mask_value: float
    type: NodeType


@dataclass
class MaxPooling1D:
    pool_size: int
    strides: int
    padding: PaddingEnum
    type: NodeType


@dataclass
class MaxPooling2D:
    pool_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    type: NodeType


@dataclass
class MaxPooling3D:
    pool_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    type: NodeType


@dataclass
class Maximum:
    type: NodeType


@dataclass
class Minimum:
    type: NodeType


@dataclass
class MultiHeadAttention:
    num_heads: int
    key_dim: int
    value_dim: List[int]
    dropout: float
    use_bias: bool
    output_shape: List[int]
    attention_axes: List[int]
    kernel_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class Multiply:
    type: NodeType


@dataclass
class Normalization:
    axis: int
    mean: List[int]
    variance: List[int]
    invert: bool
    type: NodeType


@dataclass
class PReLU:
    alpha_initializer: InitializersEnum
    alpha_regularizer: RegularizerEnum
    alpha_constraint: ConstraintEnum
    shared_axes: List[int]
    type: NodeType


@dataclass
class Permute:
    dims: List[int]
    type: NodeType


@dataclass
class RandomBrightness:
    factor: List[int]
    value_range: List[int]
    seed: List[int]
    type: NodeType


@dataclass
class RandomContrast:
    factor: List[int]
    seed: List[int]
    type: NodeType


@dataclass
class RandomCrop:
    height: int
    width: int
    seed: List[int]
    type: NodeType


@dataclass
class RandomFlip:
    mode: str
    seed: List[int]
    type: NodeType


@dataclass
class RandomFourierFeatures:
    output_dim: int
    kernel_initializer: InitializersEnum
    scale: List[int]
    trainable: bool
    type: NodeType


@dataclass
class RandomHeight:
    factor: List[int]
    interpolation: str
    seed: List[int]
    type: NodeType


@dataclass
class RandomRotation:
    factor: List[int]
    fill_mode: str
    interpolation: str
    seed: List[int]
    fill_value: float
    type: NodeType


@dataclass
class RandomTranslation:
    height_factor: List[int]
    width_factor: List[int]
    fill_mode: str
    interpolation: str
    seed: List[int]
    fill_value: float
    type: NodeType


@dataclass
class RandomWidth:
    factor: List[int]
    interpolation: str
    seed: List[int]
    type: NodeType


@dataclass
class RandomZoom:
    height_factor: List[int]
    width_factor: List[int]
    fill_mode: str
    interpolation: str
    seed: List[int]
    fill_value: float
    type: NodeType


@dataclass
class ReLU:
    max_value: List[int]
    negative_slope: float
    threshold: float
    type: NodeType


@dataclass
class RepeatVector:
    n: int
    type: NodeType


@dataclass
class Reshape:
    target_shape: List[int]
    type: NodeType


@dataclass
class Resizing:
    height: int
    width: int
    interpolation: str
    crop_to_aspect_ratio: bool
    type: NodeType


@dataclass
class SeparableConv1D:
    filters: int
    kernel_size: int
    strides: int
    padding: PaddingEnum
    dilation_rate: int
    depth_multiplier: int
    activation: ActivationsEnum
    use_bias: bool
    depthwise_initializer: InitializersEnum
    pointwise_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    depthwise_regularizer: RegularizerEnum
    pointwise_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    depthwise_constraint: ConstraintEnum
    pointwise_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class SeparableConv2D:
    filters: int
    kernel_size: List[int]
    strides: List[int]
    padding: PaddingEnum
    dilation_rate: List[int]
    depth_multiplier: int
    activation: ActivationsEnum
    use_bias: bool
    depthwise_initializer: InitializersEnum
    pointwise_initializer: InitializersEnum
    bias_initializer: InitializersEnum
    depthwise_regularizer: RegularizerEnum
    pointwise_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    depthwise_constraint: ConstraintEnum
    pointwise_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    type: NodeType


@dataclass
class SimpleRNN:
    units: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    activity_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    return_sequences: bool
    return_state: bool
    go_backwards: bool
    stateful: bool
    unroll: bool
    type: NodeType


@dataclass
class SimpleRNNCell:
    units: int
    activation: ActivationsEnum
    use_bias: bool
    kernel_initializer: InitializersEnum
    recurrent_initializer: str
    bias_initializer: InitializersEnum
    kernel_regularizer: RegularizerEnum
    recurrent_regularizer: RegularizerEnum
    bias_regularizer: RegularizerEnum
    kernel_constraint: ConstraintEnum
    recurrent_constraint: ConstraintEnum
    bias_constraint: ConstraintEnum
    dropout: float
    recurrent_dropout: float
    type: NodeType


@dataclass
class Softmax:
    axis: int
    type: NodeType


@dataclass
class SpatialDropout1D:
    rate: float
    type: NodeType


@dataclass
class SpatialDropout2D:
    rate: float
    type: NodeType


@dataclass
class SpatialDropout3D:
    rate: float
    type: NodeType


@dataclass
class StringLookup:
    max_tokens: List[int]
    num_oov_indices: int
    mask_token: List[int]
    oov_token: str
    vocabulary: List[int]
    idf_weights: List[int]
    encoding: str
    invert: bool
    output_mode: str
    sparse: bool
    pad_to_max_tokens: bool
    type: NodeType


@dataclass
class Subtract:
    type: NodeType


@dataclass
class SyncBatchNormalization:
    axis: int
    momentum: float
    epsilon: float
    center: bool
    scale: bool
    beta_initializer: InitializersEnum
    gamma_initializer: InitializersEnum
    moving_mean_initializer: InitializersEnum
    moving_variance_initializer: InitializersEnum
    beta_regularizer: RegularizerEnum
    gamma_regularizer: RegularizerEnum
    beta_constraint: ConstraintEnum
    gamma_constraint: ConstraintEnum
    type: NodeType


@dataclass
class TextVectorization:
    max_tokens: List[int]
    standardize: str
    split: str
    ngrams: List[int]
    output_mode: str
    output_sequence_length: List[int]
    pad_to_max_tokens: bool
    vocabulary: List[int]
    idf_weights: List[int]
    sparse: bool
    ragged: bool
    encoding: str
    type: NodeType


@dataclass
class ThresholdedReLU:
    theta: float
    type: NodeType


@dataclass
class TimeDistributed:
    type: NodeType


@dataclass
class UnitNormalization:
    axis: int
    type: NodeType


@dataclass
class UpSampling1D:
    size: int
    type: NodeType


@dataclass
class UpSampling2D:
    size: List[int]
    interpolation: str
    type: NodeType


@dataclass
class UpSampling3D:
    size: List[int]
    type: NodeType


@dataclass
class ZeroPadding1D:
    padding: int
    type: NodeType


@dataclass
class ZeroPadding2D:
    padding: List[int]
    type: NodeType


@dataclass
class ZeroPadding3D:
    padding: List[int]
    type: NodeType


@dataclass
class BinaryCrossentropy:
    from_logits: bool
    label_smoothing: float
    axis: int
    type: NodeType


@dataclass
class BinaryFocalCrossentropy:
    apply_class_balancing: bool
    alpha: float
    gamma: float
    from_logits: bool
    label_smoothing: float
    axis: int
    type: NodeType


@dataclass
class CategoricalCrossentropy:
    from_logits: bool
    label_smoothing: float
    axis: int
    type: NodeType


@dataclass
class CategoricalHinge:
    type: NodeType


@dataclass
class CosineSimilarity:
    axis: int
    type: NodeType


@dataclass
class Hinge:
    type: NodeType


@dataclass
class Huber:
    delta: float
    type: NodeType


@dataclass
class KLDivergence:
    type: NodeType


@dataclass
class LogCosh:
    type: NodeType


@dataclass
class MeanAbsoluteError:
    type: NodeType


@dataclass
class MeanAbsolutePercentageError:
    type: NodeType


@dataclass
class MeanSquaredError:
    type: NodeType


@dataclass
class MeanSquaredLogarithmicError:
    type: NodeType


@dataclass
class Poisson:
    type: NodeType


@dataclass
class SquaredHinge:
    type: NodeType


@dataclass
class Adadelta:
    learning_rate: float
    rho: float
    epsilon: float
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class Adagrad:
    learning_rate: float
    initial_accumulator_value: float
    epsilon: float
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class Adam:
    learning_rate: float
    beta_1: float
    beta_2: float
    epsilon: float
    amsgrad: bool
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class Adamax:
    learning_rate: float
    beta_1: float
    beta_2: float
    epsilon: float
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class Ftrl:
    learning_rate: float
    learning_rate_power: float
    initial_accumulator_value: float
    l1_regularization_strength: float
    l2_regularization_strength: float
    l2_shrinkage_regularization_strength: float
    beta: float
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class Nadam:
    learning_rate: float
    beta_1: float
    beta_2: float
    epsilon: float
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class RMSprop:
    learning_rate: float
    rho: float
    momentum: float
    epsilon: float
    centered: bool
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: int
    jit_compile: bool
    type: NodeType


@dataclass
class SGD:
    learning_rate: float
    momentum: float
    nesterov: bool
    amsgrad: bool
    weight_decay: List[int]
    clipnorm: List[int]
    clipvalue: List[int]
    global_clipnorm: List[int]
    use_ema: bool
    ema_momentum: float
    ema_overwrite_frequency: List[int]
    jit_compile: bool
    type: NodeType


@dataclass
class OnnxAbs:
    type: NodeType


@dataclass
class OnnxErf:
    type: NodeType


@dataclass
class OnnxHardSigmoid:
    alpha: float
    beta: float
    type: NodeType


@dataclass
class OnnxReduceMean:
    axes: List[int]
    keepdims: bool
    type: NodeType


@dataclass
class OnnxSqrt:
    type: NodeType


@dataclass
class Dataset:
    type: NodeType


@dataclass
class Input:
    type: NodeType


@dataclass
class RepresentationBlock:
    type: NodeType


@dataclass
class GroundTruth:
    type: NodeType


@dataclass
class CustomLayer:
    type: NodeType


@dataclass
class CustomLoss:
    type: NodeType


@dataclass
class Visualizer:
    type: NodeType


@dataclass
class Metric:
    type: NodeType


@dataclass
class Lambda:
    type: NodeType


@dataclass
class TFOpLambda:
    type: NodeType


@dataclass
class SlicingOpLambda:
    type: NodeType


@dataclass
class Repeat:
    repeats: List[int]
    axis: int
    type: NodeType


@dataclass
class Variable:
    var_shape: List[int]
    initializer: InitializerEnum
    type: NodeType


@dataclass
class Gather:
    indices: List[int]
    axis: int
    type: NodeType


@dataclass
class FixedDropout:
    rate: float
    type: NodeType


@dataclass
class NodeDataTypes:
    props: Union[Activation, ActivityRegularization, Add, AdditiveAttention, AlphaDropout, Average, AveragePooling1D, 
                 AveragePooling2D, AveragePooling3D, BatchNormalization, Bidirectional, CategoryEncoding, CenterCrop, 
                 Concatenate, Conv1D, Conv1DTranspose, Conv2D, Conv2DTranspose, Conv3D, Conv3DTranspose, ConvLSTM1D, 
                 ConvLSTM2D, ConvLSTM3D, Cropping1D, Cropping2D, Cropping3D, CuDNNGRU, CuDNNLSTM, Dense, 
                 DepthwiseConv1D, DepthwiseConv2D, Discretization, Dot, Dropout, ELU, Embedding, Flatten, GRU, GRUCell, 
                 GRUCellV1, GRUCellV2, GRUV1, GRUV2, GaussianDropout, GaussianNoise, GlobalAveragePooling1D, 
                 GlobalAveragePooling2D, GlobalAveragePooling3D, GlobalMaxPooling1D, GlobalMaxPooling2D, 
                 GlobalMaxPooling3D, GroupNormalization, IntegerLookup, LSTM, LSTMCell, LSTMCellV1, LSTMCellV2, LSTMV1, 
                 LSTMV2, LayerNormalization, LeakyReLU, LocallyConnected1D, LocallyConnected2D, Masking, MaxPooling1D, 
                 MaxPooling2D, MaxPooling3D, Maximum, Minimum, MultiHeadAttention, Multiply, Normalization, PReLU, 
                 Permute, RandomBrightness, RandomContrast, RandomCrop, RandomFlip, RandomFourierFeatures, RandomHeight, 
                 RandomRotation, RandomTranslation, RandomWidth, RandomZoom, ReLU, RepeatVector, Reshape, Resizing, 
                 SeparableConv1D, SeparableConv2D, SimpleRNN, SimpleRNNCell, Softmax, SpatialDropout1D, 
                 SpatialDropout2D, SpatialDropout3D, StringLookup, Subtract, SyncBatchNormalization, TextVectorization, 
                 ThresholdedReLU, TimeDistributed, UnitNormalization, UpSampling1D, UpSampling2D, UpSampling3D, 
                 ZeroPadding1D, ZeroPadding2D, ZeroPadding3D, BinaryCrossentropy, BinaryFocalCrossentropy, 
                 CategoricalCrossentropy, CategoricalHinge, CosineSimilarity, Hinge, Huber, KLDivergence, LogCosh, 
                 MeanAbsoluteError, MeanAbsolutePercentageError, MeanSquaredError, MeanSquaredLogarithmicError, Poisson, 
                 SquaredHinge, Adadelta, Adagrad, Adam, Adamax, Ftrl, Nadam, RMSprop, SGD, OnnxAbs, OnnxErf, 
                 OnnxHardSigmoid, OnnxReduceMean, OnnxSqrt, Dataset, Input, RepresentationBlock, GroundTruth, 
                 CustomLayer, CustomLoss, Visualizer, Metric, Lambda, TFOpLambda, SlicingOpLambda, Repeat, Variable, 
                 Gather, FixedDropout]



import json
from collections import defaultdict
from pathlib import Path
from typing import Set, Dict, Any, List, Type, Optional, Tuple
import tensorflow as tf  # type: ignore
from keras.engine.keras_tensor import KerasTensor  # type: ignore
from tensorflow.python.keras.utils.generic_utils import func_load  # type: ignore

from keras.engine.node import Node  # type: ignore
from leap_model_parser.contract.importmodelresponse import InputInfo, NodeResponse, WrapperData

layer_attributes_adaptation = {'bias_initializer', 'kernel_initializer'}

nojsonable_arg_serialization_prefix = '*tensorleap-nojsonable*'


def deserialize_non_json_arg(arg: str):
    return eval(arg[len(nojsonable_arg_serialization_prefix):])


class KerasJsonModelImport:
    def __init__(self, custom_layers: Optional[Dict[str, Type[tf.keras.layers.Layer]]] = None) -> None:
        self.id = 1
        self.nodes_cache: Dict[str, NodeResponse] = defaultdict()
        self.layer_name_to_layer: Dict[str, Dict] = defaultdict()
        self.visited_connections: Set[str] = set()
        self.ui_components: Dict[str, Dict] = defaultdict()
        self.connected_inputs: List[InputInfo] = []

        if custom_layers is None:
            custom_layers = {}
        self.custom_layers = custom_layers

    def prepare_model_to_import(self, model_schema) -> None:
        ui_components_file_path = Path(
            __file__).parent / 'contract' / 'ui_components.json'
        with open(str(ui_components_file_path), 'r') as f:
            ui_component_json = json.load(f)

        for layer in model_schema['config']['layers']:
            layer['is_representation_block'] = KerasJsonModelImport.is_representation_block(
                layer)
            self.layer_name_to_layer[layer['name']] = layer

        self.ui_components = {layer['name']: layer for layer in ui_component_json}

    @staticmethod
    def is_representation_block(layer: Dict[str, Any]) -> bool:
        return len(layer['inbound_nodes']) > 1

    def generate_graph(
            self, model_schema: Dict, layer_name_to_inbound_nodes: Dict[str, List[Node]]) -> Tuple[
        Dict[str, NodeResponse], List[InputInfo]]:
        self.prepare_model_to_import(model_schema)
        for graph_output_layer in model_schema['config']['output_layers']:
            current_layer = self.layer_name_to_layer[graph_output_layer[0]]
            self.generate_graph_recursive(
                current_layer, graph_output_layer[1], layer_name_to_inbound_nodes)

        nodes = {node.id: node for node in self.nodes_cache.values()}
        _add_origin_name_to_nodes(nodes)
        return nodes, self.connected_inputs

    def generate_graph_recursive(self, current_layer: Dict[str, Any], instance_index: int,
                                 layer_name_to_inbound_nodes: Dict[str, List[Node]]):
        self.generate_node(current_layer, instance_index,
                           layer_name_to_inbound_nodes)

        inbound_nodes = layer_name_to_inbound_nodes[current_layer['config']['name']]
        if len(inbound_nodes) == 0:
            return
        inbound_node = inbound_nodes[instance_index]
        inbound_layers = [
            parent_node.layer for parent_node in inbound_node.parent_nodes]
        for input_index, inbound_layer in enumerate(inbound_layers):
            input_id = inbound_node.flat_input_ids[input_index]
            input_node_instance_index = [node.flat_output_ids[0]
                                         for node in inbound_layer.inbound_nodes].index(input_id)
            layer_name = inbound_layer.name
            input_layer = self.layer_name_to_layer[layer_name]
            self.generate_node(
                input_layer, input_node_instance_index, layer_name_to_inbound_nodes)

            input_node_name = f"{input_layer['name']}_{input_node_instance_index}"
            current_node_name = f"{current_layer['name']}_{instance_index}"
            connection_name = f"{input_node_name}_{current_node_name}_{input_index}"
            if connection_name in self.visited_connections:
                continue

            self.visited_connections.add(connection_name)
            self.attach(input_layer, input_node_name,
                        current_node_name, input_index)
            self.generate_graph_recursive(current_layer=input_layer, instance_index=input_node_instance_index,
                                          layer_name_to_inbound_nodes=layer_name_to_inbound_nodes)

    def attach(self, input_layer: Dict, node_input_layer_name: str, node_output_layer_name: str, input_index: int):
        node_output_layer = self.nodes_cache[node_output_layer_name]
        if input_layer['class_name'] == 'InputLayer':
            if node_output_layer.name == 'Lambda':
                return
            node_input_layer_name = '-'.join(['input', input_layer['name']])

        node_input_layer = self.nodes_cache[node_input_layer_name]
        node_output_key = self.get_connection_key(
            input_layer, node_input_layer, 'outputs', input_index)
        node_input_key = self.get_connection_key(
            input_layer, node_output_layer, 'inputs', input_index)
        if node_output_key is None or node_input_key is None:
            return

        if node_output_key not in node_input_layer.outputs:
            node_input_layer.outputs[node_output_key] = {'connections': []}
        node_input_layer.outputs[node_output_key]['connections'].append({
            "node": node_output_layer.id,
            "input": node_input_key
        })

        if node_input_key not in node_output_layer.inputs:
            node_output_layer.inputs[node_input_key] = {'connections': []}
        node_output_layer.inputs[node_input_key]['connections'].append({
            "node": node_input_layer.id,
            "output": node_output_key
        })

    def _get_layer_metadata(self, layer_json: Dict) -> Optional[Dict]:
        layer_metadata = self.ui_components.get(layer_json['class_name'])
        if layer_metadata is None:
            if layer_json['class_name'] in self.custom_layers:
                layer_metadata = {
                    'type': 'CustomLayer',
                    'class_name': 'CustomLayer',
                    'selected': layer_json['class_name']
                }
        return layer_metadata

    def generate_node(
            self, layer: Dict[str, Any], instance_index: int, layer_name_to_inbound_nodes: Dict[str, List[Node]]):
        if layer['class_name'] == 'InputLayer':
            self._handle_input_layer(layer)
            return

        node_key = f"{layer['name']}_{instance_index}"
        if node_key in self.nodes_cache:
            return

        layer = self.handle_cudnn(layer)
        layer_metadata = self._get_layer_metadata(layer)
        if layer_metadata is None:
            raise Exception(
                f"{layer['class_name']} layer is not supported, please try to contact Tensorleap support")

        if layer_metadata['type'] == 'wrapper':
            wrapped_layer = self.handle_wrapper_layer(layer)
            self.generate_node(wrapped_layer, instance_index,
                               layer_name_to_inbound_nodes)
            return

        if layer['is_representation_block']:
            self.generate_rp_node(layer, layer_metadata, node_key)
        else:
            self.generate_regular_node(
                layer, layer_metadata, node_key, layer_name_to_inbound_nodes)
        self.id += 1

    @staticmethod
    def _serialize_arg(arg):
        if isinstance(arg, KerasTensor):
            return arg.name
        if hasattr(arg, 'numpy'):
            arg = arg.numpy()
        if hasattr(arg, 'tolist'):
            return arg.tolist()
        if isinstance(arg, (str, float, int, bool, type(None))):
            return arg
        if isinstance(arg, tf.dtypes.DType):
            return f'{nojsonable_arg_serialization_prefix}tf.{arg.name}'
        if isinstance(arg, (list, tuple)):
            return [KerasJsonModelImport._serialize_arg(a) for a in arg]
        if isinstance(arg, dict):
            return {k: KerasJsonModelImport._serialize_arg(v) for k, v in arg.items()}
        raise Exception('Unsupported call arg')

    @staticmethod
    def _serialize_call_args(call_args: Tuple, call_kwargs: Dict) -> str:
        serialized_call_args = [
            KerasJsonModelImport._serialize_arg(arg) for arg in call_args]
        serialized_call_kwargs = {k: KerasJsonModelImport._serialize_arg(
            v) for k, v in call_kwargs.items()}

        return json.dumps({'serialized_call_args': serialized_call_args,
                           'serialized_call_kwargs': serialized_call_kwargs})

    def generate_regular_node(self, layer: Dict[str, Any], layer_metadata: Dict[str, Any], node_key: str,
                              layer_name_to_inbound_nodes: Dict[str, List[Node]]):
        data = layer['config']
        if layer['class_name'] in ('TFOpLambda', 'SlicingOpLambda') or layer['class_name'] in self.custom_layers:
            call_args = layer_name_to_inbound_nodes[layer['config']
            ['name']][0].call_args
            call_kwargs = layer_name_to_inbound_nodes[layer['config']
            ['name']][0].call_kwargs
            data['call_args'] = self._serialize_call_args(
                call_args, call_kwargs)
        elif layer['class_name'] == "Lambda":
            lambda_func = func_load(layer['config']['function'])
            data['arg_names'] = list(lambda_func.__code__.co_varnames)

        self.layer_data_adjustments(data, layer_metadata)
        node = NodeResponse(id=str(self.id), name=layer_metadata.get(
            "class_name", layer["class_name"]), data=data)
        if 'wrapper' in layer:
            node.wrapper = layer['wrapper']
        self.nodes_cache[node_key] = node

    @classmethod
    def handle_wrapper_layer(cls, layer):
        wrapped_layer = layer['config']['layer']
        wrapped_layer['name'] = layer['name']
        wrapped_layer['is_representation_block'] = layer['is_representation_block']
        wrapper_data = layer['config'].copy()
        wrapper_data.pop('layer')
        wrapper_data.pop('name')
        # TODO: remove this hard coded condition after support
        if 'merge_mode' in wrapper_data and wrapper_data.get('merge_mode') is None:
            raise Exception(f"{layer['class_name']} layer with merge_mode None is not supported,"
                            f" please try to contact Tensorleap support")
        wrapped_layer['wrapper'] = WrapperData(
            layer['class_name'], wrapper_data)
        return wrapped_layer

    @classmethod
    def layer_data_adjustments(cls, layer_config: Dict[str, Any], layer_metadata: Dict[str, Any]) -> None:
        for layer_attribute in layer_attributes_adaptation:
            if layer_attribute in layer_config:
                layer_config[layer_attribute] = layer_config[layer_attribute]['class_name']

        layer_config["type"] = layer_metadata.get('type')
        if 'selected' in layer_metadata:
            layer_config["selected"] = layer_metadata['selected']
        # TODO: remove this hard coded condition after support
        if layer_config.get('return_state'):
            raise Exception(f"{layer_metadata.get('name')} layer with return_state True is not supported,"
                            f" please try to contact Tensorleap support")

    def generate_rp_node(self, layer: Dict[str, Any], layer_metadata: Dict[str, Any], node_key: str) -> None:
        if layer['name'] not in self.nodes_cache:
            data = layer['config']
            self.layer_data_adjustments(data, layer_metadata)
            data["output_blocks"] = []
            node = NodeResponse(id=str(self.id),
                                name=layer["class_name"], data=data)
            if 'wrapper' in layer:
                node.wrapper = layer['wrapper']
            self.nodes_cache[layer['name']] = node

            self.id += 1
        data = {
            "output_name": "Action",
            "node_id": self.nodes_cache[layer["name"]].id,
            "parent_name": layer["class_name"]
        }
        self.nodes_cache[node_key] = NodeResponse(
            id=str(self.id), name='Representation Block', data=data)
        self.nodes_cache[layer["name"]].data["output_blocks"].append({
            "block_node_id": self.nodes_cache[node_key].id,
            "output_name": "Action"
        })

    def get_connection_key(self, node_input_layer, node: NodeResponse, direction: str, index: int) -> Optional[str]:
        layer_metadata = self.ui_components.get(node.name)
        if _is_represntaion_block_node(node) or layer_metadata is None:
            if direction == 'inputs':
                connection_format = f"{node.id}-input"
            else:
                connection_format = f"{node.id}-feature_map"
        elif _is_input_node(node):
            connection_format = f"{node.id}-{node_input_layer['name']}"
        else:
            try:
                connection_format = layer_metadata[f"{direction}_data"][direction][index]['name']
            except IndexError:
                connection_format = layer_metadata[f"{direction}_data"][direction][0]['name']
            if connection_format.count('${id}') > 0:
                connection_format = f"{node.id}-{index}"
                if 'custom_input_keys' not in node.data:
                    node.data['custom_input_keys'] = []
                node.data['custom_input_keys'].append(connection_format)
            else:
                connection_format = f"{node.id}-{connection_format}"
        return connection_format

    def handle_cudnn(self, layer: Dict[str, Any]) -> Dict[str, Any]:
        if layer['class_name'] == 'CuDNNGRU':
            layer['class_name'] = 'GRU'
            gru_metadata = self.ui_components['GRU']
            gru_config = {}
            for gru_prop in gru_metadata["properties"]:
                if gru_prop["name"] in layer['config']:
                    gru_config[gru_prop["name"]
                    ] = layer['config'][gru_prop["name"]]
                elif gru_prop["isdefault"]:
                    gru_config[gru_prop["name"]] = gru_prop["default_val"]

            gru_config["name"] = layer['config']["name"]
            gru_config["trainable"] = layer['config']["trainable"]

            layer['config'] = gru_config
        return layer

    def _handle_input_layer(self, layer: Dict[str, Any]):
        input_name = layer['name']
        node_key = '-'.join(['input', input_name])
        inputNode = self.nodes_cache.get(node_key)
        if inputNode is not None:
            return
        shape = layer['config']['batch_input_shape']
        input_info = InputInfo(name=input_name, shape=shape[1:])
        self.connected_inputs.append(input_info)
        input_copy = {
            'type': 'Input',
            'output_name': input_name
        }
        inputNode = NodeResponse(
            id=str(self.id), name='Input', data=input_copy)
        self.nodes_cache[node_key] = inputNode
        self.id += 1


def _add_origin_name_to_nodes(nodes: Dict[str, NodeResponse]):
    for node in nodes.values():
        if "name" in node.data:
            node.data["origin_name"] = node.data["name"]
            node.data.pop("name")


def _is_input_node(node: NodeResponse) -> bool:
    return node.name == 'Input'


def _is_represntaion_block_node(node: NodeResponse) -> bool:
    return node.name == 'Representation Block'



from enum import Enum
from typing import Dict, Any, List, Optional

from dataclasses import dataclass, field


class ImportModelTypeEnum(Enum):
    JSON_TF2 = "JSON_TF2"
    ONNX = "ONNX"
    PB_TF2 = "PB_TF2"
    H5_TF2 = "H5_TF2"


@dataclass
class WrapperData:
    type: str
    data: Dict

@dataclass
class InputInfo:
    name: str
    shape: List[int]

@dataclass
class NodeResponse:
    id: str
    name: str
    data: Dict[str, Any] = field(default_factory=dict)
    inputs: Dict[str, Any] = field(default_factory=dict)
    outputs: Dict[str, Any] = field(default_factory=dict)
    position: List[int] = field(default_factory=lambda: [0, 0])
    wrapper: Optional[WrapperData] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['leap_model_parser',
 'leap_model_parser.contract',
 'leap_model_parser.utils',
 'leap_model_parser.utils.layerpedia',
 'leap_model_parser.utils.tlinspection',
 'leap_model_parser.utils.uicomponents']

package_data = \
{'': ['*']}

install_requires = \
['keras-data-format-converter==0.1.14',
 'leap-model-rebuilder==0.1.5',
 'numpy>=1.22.3,<2.0.0',
 'onnx2kerastl==0.0.81',
 'onnx==1.10.1']

extras_require = \
{':platform_machine == "arm64"': ['tensorflow-macos==2.11.0'],
 ':platform_machine == "x86_64"': ['tensorflow==2.11.0']}

entry_points = \
{'console_scripts': ['update_ui_components = '
                     'scripts.scripts:update_ui_components']}

setup_kwargs = {
    'name': 'leap-model-parser',
    'version': '0.1.89',
    'description': '',
    'long_description': '# Tensorleap model parser\nUsed to parse model to the import format \n',
    'author': 'idan',
    'author_email': 'idan.yogev@tensorleap.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tensorleap/leap-model-parser',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import json
import os
from typing import List, Optional

import yaml
from leap_model_parser.utils.layerpedia.layerpedia import Layerpedia
from leap_model_parser.utils.tlinspection.leapinspection import LeapInspect


class TensorFlowInspection(LeapInspect):

    def __init__(self):
        file_path = os.path.dirname(os.path.abspath(__file__))
        with open(f'{file_path}/ui_components_config.yaml') as f:
            self.yaml_conf = yaml.load(f, Loader=yaml.FullLoader)
        self.ignores = self.yaml_conf['ignores']
        self.args = self.yaml_conf['args']
        self.group_type = self.yaml_conf['group_type']

    def get_function_args(self, func):
        _args = []
        args = LeapInspect.get_function_args(func)
        for arg in args.values():
            arg_name = arg['name']
            default_val = arg['default_val']
            if arg_name in self.ignores['args_ignore']:
                continue

            is_tuple = arg.get('type') == 'tuple' and type(arg.get('default_val')) is tuple
            is_number = (arg.get('type') == 'float' or arg.get('type') == 'int') and isinstance(default_val, (int, float))
            if is_tuple or is_number:
                _args.append(arg)
                continue

            if arg_name in list(self.args['arg_group'].keys()):
                arg['type'] = 'select'
                arg['values'] = self.args['arg_values'][self.args['arg_group'][arg_name]]
                arg['isdefault'] = False
                if default_val is not None:
                    arg['isdefault'] = True
                    arg['default_val'] = default_val

            if arg_name in list(self.args['overides'].keys()):
                arg['type'] = self.args['overides'][arg_name]['type']
                arg['isdefault'] = self.args['overides'][arg_name]['isdefault']
                if arg['isdefault']:
                    override_def = self.args['overides'][arg_name]['default_val']
                    if default_val is not None and override_def != default_val:
                        raise "Override_def is not as argument default in class name: {}, argument name: {} ({} =! {})".format(
                            func, arg_name, default_val, override_def)
                    arg['default_val'] = override_def
            _args.append(arg)
        return _args

    def filter_node(self, node):
        return node['name'] in self.ignores['clsignore']

    def inspect_lib(self, lib):
        # TODO: extract only classes and not aliases
        classes = LeapInspect.get_classes(lib, self.ignores['lib_ignore'])
        nodes = []
        for node_cls in classes:
            if self.filter_node(node_cls):
                continue
            node = {
                'name': node_cls['name'],
                'parents': node_cls['parents'],
                'properties': self.get_function_args(node_cls['class']),
                'options': self.query_layerpedia_options(node_cls['name'])
            }
            shape_rank = self.query_layerpedia_shape_rank(node_cls['name'])
            if shape_rank:
                node['shape_rank'] = shape_rank
            enable_bigger_input_rank = self.query_layerpedia_enable_bigger_input_rank(
                node_cls['name'])
            if enable_bigger_input_rank:
                node['enable_bigger_input_rank'] = enable_bigger_input_rank
            template_node = {}
            for d in [*node_cls['parents'], node['name']]:
                if d in self.group_type:
                    template_node = {**template_node, **self.group_type[d]}
            node = {**template_node, **node}
            if 'hash' in node and not self.query_layerpedia_is_trainable(node_cls['name']):
                node.pop('hash')
            nodes.append(node)
        return nodes

    def inspect(self, dist: str) -> List[dict]:
        from keras import layers as tf_layers  # type: ignore
        from keras import losses as tf_losses  # type: ignore
        import tensorflow.keras.optimizers as keras_optimizers  # type: ignore
        from onnx2kerastl import customonnxlayer  # type: ignore

        libs = [tf_layers, tf_losses, keras_optimizers, customonnxlayer]
        tofile = []
        for lib in libs:
            nodes = self.inspect_lib(lib)
            tofile.extend(nodes)
        tofile.extend(self.yaml_conf['nodes'])
        with open(f'{dist}ui_components.json', 'w') as outfile:
            json.dump(tofile, outfile, indent=4)
        return tofile

    def query_layerpedia_shape_rank(self, layer_name):
        layer_knowledge = Layerpedia.get_layer_knowledge(layer_name)
        if layer_knowledge is None:
            return None
        rank = layer_knowledge.get('rank')
        return rank

    def query_layerpedia_enable_bigger_input_rank(self, layer_name) -> Optional[bool]:
        layer_knowledge = Layerpedia.get_layer_knowledge(layer_name)
        return layer_knowledge.get('enable_bigger_input_rank') if layer_knowledge else None

    def query_layerpedia_is_trainable(self, layer_name) -> Optional[bool]:
        layer_knowledge = Layerpedia.get_layer_knowledge(layer_name)
        return layer_knowledge.get('is_trainable') if layer_knowledge else None

    def query_layerpedia_options(self, layer_name):
        layer_knowledge = Layerpedia.get_layer_knowledge(layer_name)
        if layer_knowledge is None:
            return None
        layer_options = {
            "representation_block": False,
            "layer_inspection": False
        }

        if layer_knowledge['is_trainable']:
            layer_options['representation_block'] = True
            layer_options['layer_inspection'] = True

        return layer_options


if __name__ == '__main__':
    tf_inspection = TensorFlowInspection()
    tf_inspection.inspect("")
